/*****************************************************************************
*                           JMatLink                                         *
******************************************************************************
* (c) 1999 Stefan Mueller                                                    *
******************************************************************************
* 19.01.1999  beginning (reuse of HelloWorld example, Java Tutorial)  (0.01) *
* 06.02.1999  separation from gui                                     (0.02) *
* 27.03.1999  engGetScalar, engGetVector, engGetArray                 (0.03) *
* 02.05.1999  engGetScalar already works                                     *
* 20.05.1999  deleted engPutArray (float..., int...)                  (0.04) *
* 13.08.1999  thread implementation due to ActiveX thread problem     (0.05) *
* 15.08.1999  intruduced 2 locking mechanisms to lock the engine and         *
*                wait for data to be return to caller.                (0.06) *
* 16.08.1999  wait for all engine calls to return savely              (0.06a)*
* 29.08.1999  made all engine routines "synchronized"                 (0.07) *
* 11.09.1999  get char array from workspace                           (0.08) *
* 21.09.1999  engGetCharArray working (conversion to double matrix first)    *
*             (transfer as double throught engine and conversion             *
*             back to double) engine doesn't support char arrays      (0.09) *
* 10.10.1999  engPutArray(String name, double[][] values)             (0.10) *
******************************************************************************/

//****************************************************************************
// The matlab-engine must only be called from the SAME thread 
//   all the time!! On Windows systems the ActiveX implementation is
//   supposed to be the reason for this. I don't know if that happens
//   on other platforms, too.
// To achieve this. All native methods of this class, all access to
//   the engine, are called from the SAME thread. Since I don't know
//   how to call them directly I set up a mechanism to send messages
//   to that thread and ask it to process all requests. Some locking
//   mechanism make to lock up the engine for each single call in order
//   to stay out of concurrent situations.
//****************************************************************************


/* ToDo list
* engOpen is supposed to throw an exception if not successfull
* I'm not sure what happens if large matrices are passed back and
     force between this class and other applications. Maybe I
     should always clear all global arrays engGetArray1dD...
* I'm not sure if the locking mechanism is good or bad. There
     may be a problem, if two subsequent calls to two different
     engine routines occur at the same time.
*/


// **** Thread programming. Notify and start and stop ****
// Look at
// file:///e%7C/jdk1.2/docs/guide/misc/threadPrimitiveDeprecation.html
// on how to suspend, start, stop, interrupt a thread savely in JDK1.2.x
// The problem is ActiveX: one only can make calls to the engine library
// (especially engEvalString) from ONE single thread. 

import java.io.*;


public class JMatLink extends Thread{ // implements Runnable{

   // static declarations
   // the variable "status" is used to tell the main
   //   thread what to do.
   private final static int idelI            =   0;
   private final static int engOpenI         =   1;  
   private final static int engCloseI        =   2;  
   private final static int engEvalStringI   =   3;
   private final static int engGetScalarI    =   4;
   private final static int engGetVectorI    =   5;
   private final static int engGetArrayI     =   6;
   private final static int engPutArrayI     =   7;
   private final static int engPutArray1dI   =   8;
   private final static int engPutArray2dI   =   9;
   private final static int engOutputBufferI =  10;
   private final static int engGetCharArrayI =  11;  
   private final static int destroyJMatLinkI =  12;


   // All variables are global to allow all methods
   //   and the main thread to share all data
   private int              status           =   idelI;
   private String           arrayS;
   private String           engEvalStringS;
   private String           engOutputBufferS;
   private double           engGetScalarD;
   private double[]         engGetVectorD;
   private double[][]       engGetArrayD;
   private double           engPutArrayD;
   private double[]         engPutArray1dD;
   private double[][]       engPutArray2dD;
   private String[]         engGetCharArrayS;

   // Locks
   private boolean          lockEngineB       =  false;
   private boolean          lockThreadB       =  false;
   private boolean          lockWaitForValueB =  false;
   private boolean          destroyJMatLinkB  =  false;

   private Thread runner;

   // **** native declarations ****
   // NEVER call  native methods directly, like
   //   JMatLink.engEvalStringNATIVE("a=1"). Matlab's
   //   engine has quite some thread problems. 
   private native void       displayHelloWorld();
   private native void       engTestNATIVE();    

   private native void       engOpenNATIVE();
   private native void       engOpenNATIVE(String engineS);
   private native void       engCloseNATIVE();
   private native void       engEvalStringNATIVE(String evalS);
   private native double     engGetScalarNATIVE(String nameS);
   private native double[]   engGetVectorNATIVE(String nameS);
   private native double[][] engGetArrayNATIVE(String nameS);
   public  native String[]   engGetCharArrayNATIVE(String nameS);
   private native void       engPutArrayNATIVE(String matrixS, double     valuesD);
   private native void       engPutArrayNATIVE(String matrixS, double[]   valuesD);
   private native void       engPutArrayNATIVE(String matrixS, double[][] valuesD);
   private native String     engOutputBufferNATIVE();

   // **** load JMatLink library into memory ****
   static {
      try { //System.out.println("loading");
            System.loadLibrary("JMatLink");
            //System.out.println("loaded"); 
          }
      catch (UnsatisfiedLinkError e) {
          System.out.println("ERROR: Could not load the JMatLink library");
          System.out.println("       This error occures, if the path to");
          System.out.println("       matlab's <matlab>\\bin directory is");
          System.out.println("       not set properly.");
      }
   }

   // **** JMatLink constructor ****
   public JMatLink() {
      //System.out.println("JMatLink constructor");
      runner = new Thread(this);
      runner.start();
   }

   // **** terminate running thread ****
   public void destroy() {
     destroyJMatLinkB = true;
     notifyAll();
   }

   public void kill() {
     destroyJMatLinkB = true;
     callThread(destroyJMatLinkI);
   }

//////////////////////////////////////////////////////////////////////////////

   // **** engOpen ****
   // startup MATLAB and set up connection
   public synchronized void   engOpen(){
      lockEngineLock();             
      lockWaitForValue();
         callThread( engOpenI );    
      WaitForValue();
      releaseEngineLock();  
   }

   // **** engClose ****
   // close connection and terminate MATLAB
   public synchronized void   engClose(){
      lockEngineLock();
      lockWaitForValue();
         callThread( engCloseI );
      WaitForValue();
      releaseEngineLock();
   }

   // **** engEvalString ****
   // evaluate expression "evalS" in MATLAB's workspace
   public synchronized void   engEvalString(String evalS){
      //System.out.println("eval in  "+evalS);
      lockEngineLock();
      lockWaitForValue();
         engEvalStringS = evalS;
            callThread( engEvalStringI );
      WaitForValue();
      releaseEngineLock();
      //System.out.println("eval out "+evalS);
   }

  // **** engGetScalar ****
  // get scalar value or element (1,1) of an array from
  //   MATLAB's workspace
  // only real values are supported so far
  public synchronized double engGetScalar(String arrayS){
      lockEngineLock();
      lockWaitForValue();
         this.arrayS = arrayS;
         lockWaitForValue();
         callThread( engGetScalarI );
      WaitForValue();
      releaseEngineLock();
      return engGetScalarD;
  }


  // **** engGetVector ****
  // only real values are supported so far
  public synchronized double[] engGetVector(String arrayS){  
      lockEngineLock();
      lockWaitForValue();
         this.arrayS = arrayS;
         callThread( engGetVectorI );
      WaitForValue();
      releaseEngineLock();
      return engGetVectorD;

  }

  // **** engGetArray ****
  // only real values are supported so far
  public synchronized double[][] engGetArray(String arrayS){  
      lockEngineLock();
      lockWaitForValue();
         this.arrayS = arrayS;
         callThread( engGetArrayI );
      WaitForValue();
      releaseEngineLock();
      return engGetArrayD;
  }

  // **** engGetCharArray ****
  public synchronized String[] engGetCharArray(String arrayS){  
      // convert to double array
      engEvalString( "engGetCharArrayD=double(" + arrayS +")" );

      // get double array
      double[][] arrayD = engGetArray("engGetCharArrayD");

      // delete temporary double array
      engEvalString("clear engGetCharArrayD");

      // convert double back to char
      return double2String( arrayD );
  }


  // **** engPutArray ****
  // send a scalar to MATLAB
  // only real values are supported so far
  public synchronized void engPutArray(String arrayS, double     valuesD){
      lockEngineLock();
      lockWaitForValue();
         this.arrayS  = arrayS;
         engPutArrayD = valuesD;
         callThread( engPutArrayI );
      WaitForValue();
      releaseEngineLock();
  }

  // **** engPutArray ****
  // send a vector to MATLAB
  // only real values are supported so far
  public synchronized void   engPutArray(String arrayS, double[]   valuesD){
      lockEngineLock();
      lockWaitForValue();
         this.arrayS    = arrayS;
         engPutArray1dD = valuesD;
         callThread( engPutArray1dI );
      WaitForValue();
      releaseEngineLock();
  }

  // **** engPutArray ****
  // send an array to MATLAB
  // only real values are supported so far
  public synchronized void engPutArray(String arrayS, double[][] valuesD){
      lockEngineLock();
      lockWaitForValue();
         this.arrayS  = arrayS;
         engPutArray2dD = valuesD;
         callThread( engPutArray2dI );
      WaitForValue();
      releaseEngineLock();
  }


  // **** engOutputBuffer ****
  // get the output buffer from MATLAB
  public synchronized String  engOutputBuffer(){
System.out.println("Thread in: "+Thread.currentThread().getName());
      lockEngineLock();
      lockWaitForValue();
         engOutputBufferS=null;
         callThread( engOutputBufferI );
      WaitForValue();
      releaseEngineLock();
System.out.println("Thread out: "+Thread.currentThread().getName());
      return engOutputBufferS;
  }


   ////////////////////////////////////////////////////////////////
   // This method notifys the main thread to call matlab's engine
   //    Since threads don't have methods, we set a variable which
   //    contains the necessary information about what to do.
   public synchronized void callThread(int status){
      this.status = status;
      lockThreadB = false;
      notifyAll();
   }


////////////////////////////////////////////////////////////////////////////////
// The run methods does ALL calls to the native methods
// The keyword "synchronized" is neccessary to block the run()
//    method as long as one command needs to get executed.
public synchronized void run(){
   System.out.println("JMatLink: thread is running");  
   while (true) {
    // System.out.println("Anzahl der Java-Threads: "+Thread.activeCount()+"");
    // Thread thread = Thread.currentThread();
    // System.out.println("Name des aktiven Java-Threads: "+thread.getName()+"");
     //System.out.println("aktiver Java-Thread is Daemon: "+thread.isDaemon();+"");  
 

     switch (status) {
     case engOpenI:          engOpenNATIVE();
                             releaseWaitForValue();
                             break;

     case engCloseI:         engCloseNATIVE();
                             releaseWaitForValue();
                             break;

     case engEvalStringI:    engEvalStringNATIVE(engEvalStringS);
                             releaseWaitForValue();
                             break;

     case engGetScalarI:     engGetScalarD  = engGetScalarNATIVE( arrayS );
                             releaseWaitForValue();
                             break;

     case engGetVectorI:     engGetVectorD  = engGetVectorNATIVE( arrayS );
                             releaseWaitForValue();
                             break;

     case engGetArrayI:      engGetArrayD  = engGetArrayNATIVE( arrayS );
                             releaseWaitForValue();
                             break;

     case engGetCharArrayI:  engGetCharArrayS  = engGetCharArrayNATIVE( arrayS );
                             releaseWaitForValue();
                             break;

     case engPutArrayI:      engPutArrayNATIVE( arrayS, engPutArrayD );
                             releaseWaitForValue();
                             break;

     case engPutArray1dI:    engPutArrayNATIVE( arrayS, engPutArray1dD );
                             releaseWaitForValue();
                             break;

     case engPutArray2dI:    engPutArrayNATIVE( arrayS, engPutArray2dD );
                             releaseWaitForValue();
                             break;

     case engOutputBufferI:  engOutputBufferS = engOutputBufferNATIVE();
                             releaseWaitForValue();
                             break;

     default:                //System.out.println("thread default switch statem.");
     }  
     status=0;

     lockThreadB = true;
     while (lockThreadB == true) {
       synchronized(this) {
          try { wait();} // wait until next command is available
          catch (InterruptedException e) { }
       }
     }
     //System.out.println("thread awoke and passed lock"); 
     if (destroyJMatLinkB == true) break;
   } // end while
   System.out.println("thread terminated");
} // end run


////////////////////////////////////////////////////////////////////////////////
// the MATLAB engine is served by a thread. Threads don't have methods
//   which can be called. So we need to send messages to that thread
//   by using notifyAll. In the meantime NO OTHER methods is allowed to
//   access our thread (engine) so we lock everything up. 
   private void lockEngineLock(){
      synchronized(this){
         while (lockEngineB==true){
            try { //System.out.println("lockEngineLock locked");
                  wait();} // wait until last command is finished
            catch (InterruptedException e) { }
         }
         //now lockEngineB is false
         lockEngineB = true;   
      }
   } // end lockEngine

   private synchronized void releaseEngineLock(){
      lockEngineB = false;
      notifyAll();
   }

////////////////////////////////////////////////////////////////////////////////
// The MATLAB engine is served by a thread. Threads don't have methods
//    which can be called directly. If we send a command that return data
//    back to the calling function e.g. engGetArray("array"), we'll notify
//    the main thread to get the data from matlab. Since the data is collected
//    in another thread, we don't know exactly when the data is available, since
//    this is a concurrent situation. 
//    The solution is simple: I always use a locking-mechanism to wait for the
//    data. The main thread will release the lock and the calling method can
//    return the data.
//
//    Steps:
//    1. a method the returns data calls the locking method
//    2. notify the thread to call matlab
//    3. wait for the returned data
//    4. after the thread itself got the data it releases the locks method
//    5. return data

   private synchronized void lockWaitForValue(){
      lockWaitForValueB = true;
   }

   private void WaitForValue(){
      synchronized(this){
         while (lockWaitForValueB==true){
            try { //System.out.println("lockWaitForValue locked");
                  wait();} // wait for return value
            catch (InterruptedException e) { }
         }
      }
      //System.out.println("WaitForValue released");
   } // end waitForValue

   private synchronized void releaseWaitForValue(){
      lockWaitForValueB = false;
      notifyAll();
   }

  // convert an n*n double array to n*1 String vector
  public String[] double2String(double[][] d) {
    String encodeS[]=new String[d.length];  // String vector

    // for all rows
    for (int n=0; n<d.length; n++){
       byte b[] = new byte[d[n].length]; 
       // convert row from double to byte
       for (int i=0; i<d[n].length ;i++) b[i]=(byte)d[n][i];
     
       // convert byte to String
       try { encodeS[n] = new String(b, "UTF8");}
       catch (UnsupportedEncodingException e) {}
    }
    return encodeS;
  } // end double2String

} // end class JMatLink
